' LngFnam.bas
'******************************************************************************

$COMPILE        UNIT    'Destination of compilation

'Compiler directives for code generation

$CPU            80386   'Generate code for processor {8086 | 80286 | 80386}
$DIM            ALL     'Declararations required for {ALL | ARRAY | NONE}
$EVENT          OFF     'Generate event checking code
$FLOAT          EMULATE 'FP math package '{EMULATE | NPX | PROCEDURE}
$OPTIMIZE       SPEED   'Optimize code for {SPEED | SIZE}
$OPTION GOSUB   OFF     'Preserve GOSUB stack if a runtime error occurs
$OPTION SIGNED  OFF     'Return addresses by VARSEG etc. as signed integers

'Compiler directives for debugging and testing

$DEBUG UNIT     OFF     'Attach debug information to unit file
$ERROR ALL      OFF     'Generate test code for all types of errors

'******************************************************************************

'----------------------------------------------------------------------
'  Declarations
'----------------------------------------------------------------------

DEFINT A-Z

$INCLUDE        "regnames.inc"
$INCLUDE        "lngfnam.bi"

' Actions for file handling:
%FILE_CREATE                    =  &h10
%FILE_OPEN                      =     1
%FILE_TRUNCATE                  =     2

DECLARE SUB      FindCLose ()
DECLARE FUNCTION GetDriveNumber (Drive AS STRING) AS BYTE
DECLARE FUNCTION SetOnExit (BYVAL ProcPtr AS DWORD) AS INTEGER

'----------------------------------------------------------------------
' Shared variables
'----------------------------------------------------------------------

DIM LfnEnabled  AS SHARED INTEGER       ' flag telling if Long File Names
                                        ' are enabled already or not
DIM DOSError    AS SHARED INTEGER       ' DOS error code

DIM FindHandle  AS SHARED WORD          ' Handle for Win32 file info
DIM FindData    AS SHARED Win32_Find_Data       ' Structure for ditto
DIM PtrFindData AS SHARED Win32_Find_Data PTR   ' and its pointer

DIM DTA         AS SHARED FileInfo PTR  ' pointer to structure for DOS
					' file info

'**********************************************************************

$CODE SEG       "lngfnam_text"

'----------------------------------------------------------------------
' Int21: the DOS interrupt procedure
'----------------------------------------------------------------------
OldInt21:
ASM     DD      0

Int21:

ASM     PUSHF                           ; save caller's flags
ASM     STI                             ; enable interrupts
ASM
ASM     CMP     AH, &h71                ; Win95 long file name functions
ASM     JE      I21_Jmp                 ; directly forwarded to DOS

' N.B. The functions
'       7147h (get active directory) and
'       7160h (get full path name)
' are not included, because the caller generally will have declared a
' buffer too small for Windows 95 long file names

I21_039:

ASM     CMP     AH, &h39                ; make subdirectory
ASM     JNZ     I21_L1
ASM     PUSH    SI
ASM     MOV     SI, DX
ASM     CALL    NEAR PTR ResetCase      ; reset case of dir name handled
ASM                                     ; with PreCase$
ASM     POP     SI
ASM     JMP     SHORT I21_071

I21_L1:

ASM     CMP     AH, &h3A                ; erase subdirectory
ASM     JZ      I21_071
ASM
ASM     CMP     AH, &h3B                ; set actual directory
ASM     JZ      I21_071
ASM
ASM     CMP     AH, &h56                ; rename file
ASM     JNZ     I21_03C

I21_071:

ASM     MOV     AL, AH
ASM     MOV     AH, &h71

I21_Jmp:

ASM     POPF                            ; restore caller's flags
ASM     CLI                             ; disable interrupts
ASM     JMP     DWORD PTR OldInt21

I21_03C:

ASM     CMP     AH, &h3C                ; create file
ASM                                     ; overwrites existing file
ASM     JE      I21_05B
ASM     CMP     AH, &h3D                ; open file
ASM     JE      I21_05B
ASM     CMP     AH, &h5B                ; create file
ASM                                     ; fails if file exists already
ASM     JNZ     I21_06C

I21_05B:

ASM     POPF                            ; restore flags
ASM
ASM     PUSH    BX                      ; preserve unused registers
ASM     PUSH    CX
ASM     PUSH    DX
ASM     PUSH    DI
ASM     PUSH    SI
ASM     PUSHF                           ; flags on stack for call of ISR
ASM
ASM     ;       DS                      ; segment file name is given
ASM     MOV     SI, DX                  ; offset file name
ASM     CALL    NEAR PTR ResetCase      ; reset case of filename handled
ASM                                     ; with PreCase$
ASM     XOR     DI, DI                  ; no alias hint
ASM
ASM     CMP     AH, &h3D
ASM     JE      I21_03D
ASM
ASM     ; Create file
ASM
ASM     MOV     BX, %OPEN_ACCESS_READWRITE + %OPEN_SHARE_COMPATIBLE
ASM                                     ; access & sharing modes & flags
ASM     ;       CX                      ; file attribute is given
ASM
ASM     CMP     AH, &h3C
ASM     JNE     I21_L01
ASM     MOV     DX, %FILE_CREATE + %FILE_TRUNCATE
ASM                                     ; function 3Ch's action
ASM     JMP     SHORT I21_OpenCreate

I21_L01:

ASM     MOV     DX, %FILE_CREATE        ; function 5Bh's action
ASM     JMP     SHORT I21_OpenCreate

I21_03D:        ' Open file

ASM     MOV     BL, AL                  ; access & sharing modes
ASM     XOR     BH, BH                  ; flags
ASM     MOV     CX, %FILE_ATTRIBUTE_ALL
ASM                                     ; opens all files
ASM     MOV     DX, %FILE_OPEN          ; function 3Dh's action

I21_OpenCreate: ' Execute interrupt

ASM     MOV     AX, &h716C              ; Function
ASM                                     ; Flags are on stack already
ASM     CLI                             ; Disable interrupts
ASM     CALL    DWORD PTR OldInt21      ; Call original ISR
ASM
ASM     POP     SI                      ; restore registers
ASM     POP     DI
ASM     POP     DX
ASM     POP     CX
ASM     POP     BX
ASM
ASM     RETF    2                       ; return from interrupt with
ASM                                     ; current flags

I21_06C:

ASM     CMP     AH, &h6C                ; file open/create
ASM     JNZ     I21_041
ASM
ASM     POPF                            ; restore flags
ASM     PUSH    DI                      ; preserve unused registers
ASM     PUSHF                           ; flags on stack for call of ISR
ASM     CALL    NEAR PTR ResetCase      ; reset case of filename handled
ASM                                     ; with PreCase$
ASM     XOR     DI, DI                  ; no alias hint
ASM     MOV     AX, &h716C              ; function
ASM     CLI                             ; disable interrupts
ASM     CALL    DWORD PTR OldInt21      ; call original ISR
ASM     POP     DI
ASM     RETF    2                       ; return from interrupt with
ASM                                     ; current flags

I21_041:

ASM     CMP     AH, &h41                ; erase file
ASM     JNZ     I21_043
ASM
ASM     POPF                            ; restore flags
ASM     PUSH    CX                      ; preserve unused registers
ASM     PUSH    SI
ASM     PUSHF                           ; flags on stack for call of ISR
ASM     MOV     CH, %FILE_ATTRIBUTE_NORMAL
ASM                                     ; matches all files
ASM     MOV     CL, %FILE_ATTRIBUTE_READWRITE
ASM                                     ; doesn't erase read-only files
ASM     XOR     SI, SI                  ; wildcards not permitted
ASM     MOV     AX, &h7141              ; function
ASM     CLI                             ; disable interrupts
ASM     CALL    DWORD PTR OldInt21      ; call original ISR
ASM     POP     SI
ASM     POP     CX
ASM     RETF    2                       ; return from interrupt with
ASM                                     ; current flags

I21_043:

ASM     CMP     AH, &h43                ; get or set file attributes
ASM     JNE     I21_Jmp2
ASM
ASM     POPF                            ; restore flags
ASM     PUSH    CX                      ; preserve unused registers
ASM     PUSH    DI
ASM     PUSH    SI
ASM     PUSHF                           ; flags on stack for call of ISR
ASM
ASM     AND     AL, 1                   ; set or get attributes only
ASM     MOV     BL, AL                  ; subfunction
ASM     ;       CX                      ; attributes are given
ASM     ;       DX                      ; offset file name is given
ASM     ;       DS                      ; segment file name is given
ASM     MOV     AX, &h7143              ; function
ASM
ASM     CLI                             ; disable interrupts
ASM     CALL    DWORD PTR OldInt21      ; call original ISR
ASM
ASM     POP     SI
ASM     POP     DI
ASM     POP     BX
ASM
ASM     RETF    2                       ; return from interrupt with
ASM                                     ; current flags

I21_Jmp2:

ASM     JMP     I21_Jmp

'----------------------------------------------------------------------

ResetCase:
' A file name or directory name can be handled by PreCase$ before
' creating the file with the OPEN statement or the directory with the
' MKDIR statement. PreCase$ sets the highest bit of the ASCII code of
' all lower case letters to 1 in order to prevent PowerBasic from
' capitalize them. When executing the DOS interrupt, ResetCase resets
' the highest bit to 0, so DOS gets the filename with case preserved.

ASM     PUSH    AX
ASM     PUSH    ES
ASM     PUSH    SI
ASM     PUSH    DI

ASM     MOV     AX, DS
ASM     MOV     ES, AX
ASM     MOV     DI, SI

I21_L5:

ASM     LODSB                           ; get ASCII code
ASM     CMP     AL, 0                   ; is it zero?
ASM     JE      I21_L7                  ; then end of string
ASM     CMP     AL, &hE1                ; If ASCII < &hE1
ASM     JB      I21_L6                  ; or
ASM     CMP     AL, &hFA                ; if ASCII > &hFA
ASM     JA      I21_L6                  ; it's no lower case letter
ASM     AND     AL, &h7F                ; Unset highest bit of lower case letter

I21_L6:

ASM     STOSB                           ; store changed ASCII code
ASM     JMP     I21_L5                  ; and continue with next character

I21_L7:

ASM     POP     DI
ASM     POP     SI
ASM     POP     ES
ASM     POP     AX
ASM     RETN

'**********************************************************************

FUNCTION EnableLfn () PUBLIC AS INTEGER
'----------------------------------------------------------------------
' EnableLfn enables the use of long file names with PowerBasic
' statements
'----------------------------------------------------------------------
DIM Address     AS WORD PTR

IF LfnEnabled THEN

  FUNCTION = %TRUE

ELSEIF NOT LfnSupported THEN

  ' Test for support of Long File Names

  FUNCTION = %FALSE

ELSE

  ' Save current Int 21 vector in OldInt21

  REG %AX, &H03521
  CALL INTERRUPT &H021
  Address = CODEPTR32(OldInt21)
  @Address = REG(%BX)
  INCR Address, 2
  @Address = REG(%ES)

  ' Give new Int 21 vector Int21 to DOS

  REG %DS, CODESEG(Int21)
  REG %DX, CODEPTR(Int21)
  REG %AX, &H02521
  CALL INTERRUPT &H021

  ' Set own interrupt flag

  LfnEnabled = %TRUE
  FUNCTION = %TRUE

  ' DisableLfn has to be called when the program terminates

  SetOnExit CODEPTR32(DisableLfn)

END IF
END FUNCTION

'**********************************************************************

SUB DisableLfn () PUBLIC
'----------------------------------------------------------------------
' DisableLfn disables the use of long file names with PowerBasic
' statements
'----------------------------------------------------------------------

DIM Address     AS WORD PTR

IF LfnEnabled THEN

  ' Restore original interrupt vector

  Address = CODEPTR32(OldInt21)
  REG %DX, @Address
  INCR Address, 2
  REG %DS, @Address
  REG %AX, &H02521
  CALL INTERRUPT &H021

  ' Lower own interrupt flag

  LfnEnabled = %FALSE

END IF
END SUB

'**********************************************************************

FUNCTION LfnSupported () PUBLIC AS INTEGER
'----------------------------------------------------------------------
' LfnSupported returns %TRUE (-1) if Windows 95 long file names are
' supported, and %FALSE (0) otherwise
'----------------------------------------------------------------------

DIM DOSVersion  AS BYTE                 ' main DOS version number
DIM LfnSupport  AS STATIC INTEGER       ' flag telling if Long File Names
                                        ' are supported (1 = supported,
                                        ' -1 = not supported)
DIM SystemName  AS STRING
DIM RootName    AS STRING
DIM LenSysName  AS INTEGER
DIM PtrSysName  AS STRING PTR
DIM PtrRootName AS STRING PTR

' First test if function was called already.
' If so, result is in flag LfnSupport

IF LfnSupport THEN

  FUNCTION = (LfnSupport > 0)

ELSE

  'Initialize function value and flag

  FUNCTION = %FALSE
  LfnSupport = -1

  'Do we have DOS 7 or higher (but not a DOS box under OS/2)?
  'N.B.: OS/2 version 2's DOS box has version number 20

  ASM   XOR     BX, BX                  ;Zero BX
  ASM   MOV     AX, &h3306              ;get DOS version (5 or higher)
  ASM   INT     &h21
  ASM   MOV     DOSVersion, BL

  IF DOSVersion < 7 OR DOSVersion >= 20 THEN

    ' If DOSVersion is 5 or 6 or 0 (DOS 4 or lower), or 20 or higher
    ' (OS/2), then no long Win95 names

    EXIT FUNCTION
  END IF

  'Do we have MS-DOS 7?

  ASM   PUSH    DS
  ASM   MOV     AX, &h4A33              ;function 4A33h
  ASM   INT     &h2F                    ;of ISR 2Fh
  ASM   POP     DS
  ASM   CMP     AX, 0                   ;is AX zeroed?
  ASM   JE      L2                      ;if not, no MS-DOS 7
    EXIT FUNCTION
  L2:

  'Does system support long names?

  SystemName$ = STRING$(8,0)
  PtrSysName  = STRPTR32(SystemName$)
  LenSysName  = LEN(SystemName$)
  RootName$   = "C:\" + CHR$(0)
  PtrRootName = STRPTR32(RootName$)

  ASM   PUSH    DS
  ASM   PUSH    DI
  ASM   LES     DI, PtrSysName          ;get address of SystemName in ES:DI
  ASM   MOV     CX, LenSysName          ;   and length in CX
  ASM   LDS     DX, PtrRootName         ;get address of RootName in DS:DX
  ASM   MOV     AX, &h71A0              ;get volume information
  ASM   INT     &h21
  ASM   POP     DI
  ASM   POP     DS
  ASM   JC      L3                      ;if OK,
    FUNCTION = %TRUE                    'then long names are supported
    LfnSupport = 1
  L3:

END IF
END FUNCTION

'**********************************************************************

FUNCTION PreCase (File AS STRING) PUBLIC AS STRING
'----------------------------------------------------------------------
' Function:
' A file name or directory name can be handled by PreCase$ before
' creating the file with the OPEN statement or the directory with the
' MKDIR statement. PreCase$ sets the highest bit of the ASCII code of
' all lower case letters to 1 in order to prevent PowerBasic from
' capitalize them. When executing the DOS interrupt, ResetCase resets
' the highest bit to 0, so DOS gets the filename with case preserved.

' If long file names aren't enabled PreCase$ returns the file name
' undisturbed.

' Examples:
' OPEN PreCase(File$) FOR OUTPUT AS 1
' MKDIR PreCase(Directory$)
'----------------------------------------------------------------------

DIM Address     AS STRING PTR
DIM Temp        AS STRING
DIM Length      AS INTEGER

IF LfnEnabled THEN

  Temp = File
  Address = STRPTR32(Temp)
  Length = LEN(Temp)

  ASM   PUSH    DS
  ASM   PUSH    SI
  ASM   PUSH    DI
  ASM   LES     DI, Address             ; get address in ES:DI
  ASM   LDS     SI, Address             ;   and in DS:SI
  ASM   MOV     CX, Length              ;   and length in CX

  L4:

  ASM   LODSB                           ; get ASCII code
  ASM   CMP     AL, &h61                ; If ASCII < &h61
  ASM   JB      L5                      ; or
  ASM   CMP     AL, &h7A                ; if ASCII > &h7a
  ASM   JA      L5                      ; it's no lower case letter
  ASM   OR      AL, &h80                ; Set highest bit of lower case letter

  L5:

  ASM   STOSB                           ; store changed ASCII code
  ASM   LOOP    L4

  ASM   POP     DI
  ASM   POP     SI
  ASM   POP     DS

  FUNCTION = Temp

ELSE
  FUNCTION = File
END IF

END FUNCTION

'**********************************************************************

FUNCTION DOSErr () PUBLIC AS INTEGER
'----------------------------------------------------------------------
' Returns the DOS error code
'----------------------------------------------------------------------
  FUNCTION = DOSError
END FUNCTION

FUNCTION DOSErrTest () PUBLIC AS INTEGER
'----------------------------------------------------------------------
' Returns the DOS error code, and resets it to zero
'----------------------------------------------------------------------
  FUNCTION = DOSError
  DOSError = 0
END FUNCTION

'**********************************************************************

FUNCTION LfnKill (BYVAL File AS STRING) PUBLIC AS INTEGER
'----------------------------------------------------------------------
' Long file name replacement for PowerBasic's KILL statement.
' The function returns %TRUE (-1) if all went well, and %FALSE (0)
' otherwise.
' Of course long file names can be used under Windows 95 only.
'----------------------------------------------------------------------

DIM Address     AS STRING PTR

File = File + CHR$(0)
Address = STRPTR32(File)

ASM     PUSH    DS
ASM     PUSH    SI
IF LfnSupported THEN                    ' Long file names supported ?
  ASM   MOV     SI, 1                   ; wildcards allowed
  ASM   MOV     CH, %FILE_ATTRIBUTE_NORMAL      ; must match attrib's
  ASM   MOV     CL, %FILE_ATTRIBUTE_NORMAL      ; search attrib's
  ASM   MOV     AX, &h7141              ; Win95 function delete file
ELSE
  ASM   MOV     AH, &h41                ; DOS function delete file
END IF
ASM     LDS     DX, Address             ; address of file name
ASM     INT     &h21
ASM     POP     SI
ASM     POP     DS
ASM     JC      L10                     ; if error carry set and error no in AX
ASM     XOR     AX, AX                  ; otherwise AX = 0
L10:
ASM     MOV     DOSError, AX            ; DOS error code

FUNCTION = (DOSError = 0)               ' return value

END FUNCTION

'**********************************************************************

FUNCTION LfnAttrib (BYVAL File AS STRING) PUBLIC AS INTEGER
'----------------------------------------------------------------------
' Long file name replacement for PowerBasic's ATTRIB function.
' If an error is encountered the function returns -1 as its value.
' Of course long file names can be used under Windows 95 only.
'----------------------------------------------------------------------

DIM Address     AS STRING PTR
DIM Attribs     AS INTEGER

FUNCTION = -1				' Initialize function value
File = File + CHR$(0)
Address = STRPTR32(File)

IF LfnSupported THEN                    ' Long file names supported ?
  ASM   MOV     AX, &h7143              ; Win95 function get or set attributes
  ASM   MOV     BL, 0                   ; action: retrieve attributes
ELSE
  ASM   MOV     AH, &h43                ; DOS function get or set attributes
  ASM   MOV     AL, 0                   ; action: retrieve attributes
END IF
ASM     PUSH    DS
ASM     LDS     DX, Address             ; address of file name
ASM     INT     &h21
ASM     POP     DS
ASM     JC      L11                     ; if error carry set and error no in AX
ASM     XOR     AX, AX                  ; otherwise AX = 0
ASM     MOV     FUNCTION, CX
L11:
ASM     MOV     DOSError, AX                ; DOS error code in Temp

END FUNCTION

'**********************************************************************

FUNCTION LfnName (BYVAL OldName AS STRING, BYVAL NewName AS STRING) PUBLIC _
         AS INTEGER
'----------------------------------------------------------------------
' Long file name replacement for PowerBasic's NAME statement.
' The function returns %TRUE (-1) if all went well, and %FALSE (0)
' otherwise.
' Of course long file names can be used under Windows 95 only.
'----------------------------------------------------------------------

DIM OldAddr     AS STRING PTR
DIM NewAddr     AS STRING PTR

OldName = OldName + CHR$(0)
NewName = NewName + CHR$(0)
OldAddr = STRPTR32(OldName)
NewAddr = STRPTR32(NewName)

IF LfnSupported THEN                    ' Long file names supported ?
  ASM   MOV     AX, &h7156              ; Win95 function rename file
ELSE
  ASM   MOV     AH, &h56                ; DOS function rename file
END IF
ASM     PUSH    DS
ASM     PUSH    DI
ASM     LES     DI, NewAddr             ; address of new name
ASM     LDS     DX, OldAddr             ; address of old name
ASM     INT     &h21
ASM     POP     DI
ASM     POP     DS
ASM     JC      L12                     ; if error carry set and error no in AX
ASM     XOR     AX, AX                  ; otherwise AX = 0
L12:
ASM     MOV     DOSError, AX            ; DOS error code

FUNCTION = (DOSError = 0)               ' return value

END FUNCTION

'**********************************************************************

FUNCTION LfnCurDir (BYVAL Drive AS STRING) PUBLIC AS STRING
'----------------------------------------------------------------------
' Long file name replacement for PowerBasic's CURDIR function.
' If an error is encountered the function returns "" as its value.
' Win95 returns the path all in caps.
'----------------------------------------------------------------------

DIM Buffer      AS STRING
DIM BufAddr     AS STRING PTR
DIM BufLen      AS INTEGER
DIM DriveNumber AS BYTE
DIM Lfn         AS INTEGER

IF LEN(Drive) = 0 THEN
  Drive = CurDrive
ELSE
  Drive = UCASE$(LEFT$(Drive$, 1))
END IF
DriveNumber = GetDriveNumber(Drive)
Lfn = LfnSupported
IF Lfn THEN
  BufLen = %MAX_PATH
ELSE
  BufLen = %MAX_DOS_PATH
END IF
Buffer = STRING$(BufLen, 0)
BufAddr = STRPTR32(Buffer)

IF Lfn THEN                             ' Long file names supported ?
  ASM   MOV     AX, &h7147              ; Win95 function get current dir
ELSE
  ASM   MOV     AH, &h47                ; DOS function get current dir
END IF
ASM     PUSH    DS
ASM     PUSH    SI
ASM     LDS     SI, BufAddr             ; address of buffer in DS:SI
ASM     XOR     DH, DH
ASM     MOV     DL, DriveNumber         ; drive number in DX
ASM     INT     &h21
ASM     POP     SI
ASM     POP     DS
ASM     JC      L13                     ; if error carry set and error no in AX
ASM     XOR     AX, AX                  ; otherwise AX = 0
L13:
ASM     MOV     DOSError, AX            ; DOS error code

IF DOSError THEN
  FUNCTION = ""
ELSE
  FUNCTION = Drive + ":\" + RTRIM$(Buffer, CHR$(0)) ' the directory path
END IF

END FUNCTION

'**********************************************************************

FUNCTION CurDrive () PUBLIC AS STRING
DIM DriveNumber AS BYTE

ASM     MOV     AH, &h19                ; get current disk drive
ASM     INT     &h21
ASM     INC     AL
ASM     MOV     DriveNumber, AL

FUNCTION = CHR$(DriveNumber + 64)
END FUNCTION

'**********************************************************************

FUNCTION LfnDirFirst (BYVAL Mask AS STRING, BYVAL Attr AS BYTE) PUBLIC _
         AS STRING
'----------------------------------------------------------------------
' Long file name replacement for PowerBasic's DIR function, first time.
' If an error is encountered the function returns "" as its value.
'----------------------------------------------------------------------

DIM PtrMask     AS STRING PTR

Mask = Mask + CHR$(0)
PtrMask = STRPTR32(Mask)

IF LfnSupported THEN

  ' Close a search handle not yet closed

  IF FindHandle THEN
    FindClose
  END IF

  PtrFindData = VARPTR32(FindData)

  ' Clean the data structure

  FindData.cFileName = STRING$(%MAX_PATH, 0)
  FindData.cAlternateFileName = STRING$(14, 0)

  ASM   MOV     AX, &h714E
  ASM   MOV     CL, Attr                        ; search attributes
  ASM   TEST    CL, %FILE_ATTRIBUTE_VOLUME      ; volume attribute included?
  ASM   JZ      L20                             ; then

  ASM   MOV     CH, %FILE_ATTRIBUTE_VOLUME      ; only volume names match
  ASM   JMP     SHORT L21

  L20:                                          ' otherwise

  ASM   OR      CL, %FILE_ATTRIBUTE_READONLY    ; include read-only files
  ASM                                           ; as does PB's DIR$
  ASM   MOV     CH, %FILE_ATTRIBUTE_NORMAL      ; and all files match

  L21:

  ASM   PUSH    DS                      ; preserve registers
  ASM   PUSH    SI
  ASM   PUSH    DI
  ASM   MOV     SI, 1                   ; MS-DOS date- and time format
  ASM   LES     DI, PtrFindData         ; Address of Win32_Find_Data structure
  ASM                                   ; receiving the file info
  ASM   LDS     DX, PtrMask             ; address of file mask
  ASM   INT     &h21
  ASM   POP     DI
  ASM   POP     SI
  ASM   POP     DS
  ASM   JC      L22                     ; everything OK?
  ASM   MOV     FindHandle, AX          ; then save search handle
  ASM   XOR     AX, AX                  ; no error

  L22:

  ASM   MOV     DOSError, AX            ; DOS error code

  IF DOSError THEN
    FindCLose                           ' close the find procedure
    FUNCTION = ""
  ELSE
    FUNCTION = RTRIM$(FindData.cFileName, CHR$(0))    ' the file name
  END IF

ELSE
  GetDTA  				' Get addres of DTA
  @DTA.fiFileName = STRING$(13, 0)	' Clean the data structure

  ASM   MOV     AH, &h4E
  ASM   MOV     CL, Attr                        ; search attributes
  ASM   OR      CL, %FILE_ATTRIBUTE_READONLY    ; include read-only files
  ASM                                           ; as does PB's DIR$
  ASM   XOR     CH, CH

  ASM   PUSH    DS                      ; preserve registers
  ASM   LDS     DX, PtrMask             ; address of file mask
  ASM   INT     &h21
  ASM   POP     DS
  ASM   JC      L23                     ; everything OK?
  ASM   XOR     AX, AX                  ; no error

  L23:

  ASM   MOV     DOSError, AX            ; DOS error code

  IF DOSError THEN
    FUNCTION = ""
  ELSE
    FUNCTION = RTRIM$(@DTA.fiFileName, CHR$(0))    ' the file name
  END IF

END IF

END FUNCTION

'**********************************************************************

FUNCTION LfnDirNext () PUBLIC AS STRING
'----------------------------------------------------------------------
' Long file name replacement for PowerBasic's DIR function, next times.
' If an error is encountered the function returns "" as its value.
'----------------------------------------------------------------------

IF LfnSupported THEN
  ' Clean the data structure

  FindData.cFileName = STRING$(%MAX_PATH, 0)
  FindData.cAlternateFileName = STRING$(14, 0)

  ASM   MOV     AX, &h714F
  ASM   MOV     BX, FindHandle
  ASM   PUSH    SI                      ; preserve registers
  ASM   PUSH    DI
  ASM   MOV     SI, 1                   ; MS-DOS date- and time format
  ASM   LES     DI, PtrFindData         ; Address of Win32_Find_Data structure
  ASM                                   ; receiving the file info
  ASM   INT     &h21
  ASM   POP     DI
  ASM   POP     SI
  ASM   JC      L24                     ; everything OK?
  ASM   XOR     AX, AX                  ; signal no error

  L24:                                  ' if error

  ASM   MOV     DOSError, AX            ; DOS error code

  IF DOSError THEN
    FindCLose                           ' close the find procedure
    FUNCTION = ""
  ELSE
    FUNCTION = RTRIM$(FindData.cFileName, CHR$(0))    ' the file name
  END IF
ELSE
  GetDTA  				' Get addres of DTA
  @DTA.fiFileName = STRING$(13, 0)	' Clean the data structure

  ASM   MOV     AH, &h4F
  ASM   INT     &h21
  ASM   JC      L25                     ; everything OK?
  ASM   XOR     AX, AX                  ; no error

  L25:

  ASM   MOV     DOSError, AX            ; DOS error code

  IF DOSError THEN
    FUNCTION = ""
  ELSE
    FUNCTION = RTRIM$(@DTA.fiFileName, CHR$(0))    ' the file name
  END IF

END IF
END FUNCTION

'**********************************************************************

FUNCTION LfnDirAlias () PUBLIC AS STRING
'----------------------------------------------------------------------
' Returns the alternate filename (alias) after a call to LfnDirFirst or
' LfnDirNext has been made.
' On it self it doesn't return a meaningfull result.
'----------------------------------------------------------------------
IF LfnSupported AND FindHandle <> 0 THEN
  FUNCTION = RTRIM$(FindData.cAlternateFileName, CHR$(0))
ELSE
  FUNCTION = ""
END IF
END FUNCTION

'**********************************************************************

SUB FindCLose () PRIVATE
'----------------------------------------------------------------------
' Closes the long filename find procedure and frees its internal storage.
'----------------------------------------------------------------------

ASM     MOV     AX, &h71A1              ; close find procedure
ASM     MOV     BX, FindHandle
ASM     INT     &h21

FindHandle = 0

END SUB

'**********************************************************************

FUNCTION GetPtrFindData () PUBLIC AS DWORD
'----------------------------------------------------------------------
' Returns a pointer to the Win32 FindData structure
'----------------------------------------------------------------------
GetPtrFindData = PtrFindData
END FUNCTION

'**********************************************************************

FUNCTION GetDTA () PUBLIC AS DWORD
'----------------------------------------------------------------------
' Returns a pointer to the DOS FileInfo structure
'----------------------------------------------------------------------
ASM     MOV     AH, &h2F
ASM     INT     &h21
ASM     MOV     WORD PTR DTA, BX
ASM     MOV     WORD PTR DTA[2], ES
FUNCTION = DTA
END FUNCTION

'**********************************************************************

SUB LfnFiles () PUBLIC
'----------------------------------------------------------------------
' Long file name replacement for PowerBasic's FILES statement
'----------------------------------------------------------------------
DIM col		AS INTEGER
DIM File	AS STRING

IF lfnSupported THEN
  File$ = LfnDirFirst$("*.*", &h10)
  DO WHILE LEN(File$)
    PRINT LfnDirAlias$;
    IF (LfnAttrib(File$) AND &h10) THEN
      PRINT TAB(14); "<DIR>";
    END IF
    PRINT TAB(22); File$
    File$ = LfnDirNext$
  LOOP
ELSE
  col = 1
  File$ = LfnDirFirst$("*.*", &h10)
  DO WHILE LEN(File$)
    IF col > 80 THEN
      PRINT
      col = 1
    END IF
    PRINT TAB(col); File$;
    IF (LfnAttrib(File$) AND &h10) THEN
      PRINT TAB(col + 13); "<DIR>";
    END IF
    INCR col, 20
    File$ = LfnDirNext$
  LOOP
  PRINT
END IF
PRINT USING$(STRING$(10, "#"), DiskFree("")); " Bytes free"

END SUB

'**********************************************************************

FUNCTION DiskFree (BYVAL Drive AS STRING) PUBLIC AS DWORD
'----------------------------------------------------------------------
' Returns the number of free bytes on the given drive.
' If the given drive is invalid, the function returns 0, and DOSError
' has the value 15 (invalid disk drive).
'----------------------------------------------------------------------

DIM DriveNumber         AS BYTE
DIM BytesInSector       AS WORD
DIM SectorsInCluster    AS WORD
DIM FreeClusters        AS WORD

DriveNumber = GetDriveNumber(Drive)

ASM     MOV     AH, &h36
ASM     MOV     DL, DriveNumber
ASM     INT     &h21
ASM     CMP     AX, -1                  ; invalid disk drive if AX = -1
ASM     JE      L30
ASM     MOV     BytesInSector, CX
ASM     MOV     SectorsInCluster, AX
ASM     MOV     FreeClusters, BX

FUNCTION = CDWD(FreeClusters) * SectorsInCluster * BytesInSector
DOSError = 0
EXIT FUNCTION

L30:
FUNCTION = 0
DOSError = 15                           ' invalid disk drive

END FUNCTION

'**********************************************************************

FUNCTION GetDriveNumber (Drive AS STRING) PRIVATE AS BYTE
'----------------------------------------------------------------------
' Sets drive number for given drive:
' A or a = 1, B or b = 2 etc.
' If Drive is an empty string the drive number is 0 to designate the
' current drive. If the first character of Drive isn't a letter, the
' drive number is 255 to designate an invalid drive.
'----------------------------------------------------------------------
DIM DriveNumber AS INTEGER
DriveNumber = ASCII(Drive)
SELECT CASE DriveNumber
CASE -1
  DriveNumber = 0
CASE 65 TO 90
  DECR DriveNumber, 64
CASE 97 TO 122
  DECR DriveNumber, 96
CASE ELSE
  DriveNumber = 255
END SELECT
FUNCTION = DriveNumber
END FUNCTION
